/* A repository that contains information about a package.

Metropolis: Design Environment for Heterogeneus Systems.

Copyright (c) 1998-2005 The Regents of the University of California.
All rights reserved.

Permission is hereby granted, without written agreement and without
license or royalty fees, to use, copy, modify, and distribute this
software and its documentation for any purpose, provided that the
above copyright notice and the following two paragraphs appear in all
copies of this software and that appropriate acknowledgments are made
to the research of the Metropolis group.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS.


METROPOLIS_COPYRIGHT_VERSION_1
COPYRIGHTENDKEY


*/


package metropolis.metamodel.frontend;

import metropolis.metamodel.Scope;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;

//////////////////////////////////////////////////////////////////////////
//// PackageRepository
/**
 * A repository that contains information about a package. This information
 * is sufficient to know the members of a package without specifically
 * having to load and parse all of them.
 * <p>
 * This repository also contains information about the source file where
 * a ObjectDecl can be found, and the list of ObjectDecls that can be
 * found in each file. This information can be used to recover the
 * information of the repository incrementally, in case that some files
 * in the package have been modified, added or removed.
 * <p>
 * Portions of this code were derived from sources developed under the
 * auspices of the Ptolemy II project.
 *
 * @author   Robert Clariso
 * @version  $Id: PackageRepository.java,v 1.15.4.1 2005/02/14 18:46:13 cxh Exp $
 */
public class PackageRepository extends DirRepository {

    ///////////////////////////////////////////////////////////////////
    ////                           constructors                    ////

    /** Create a new repository for a package in a given directory. The
     *  contents of the repository will have to be loaded using the
     *  method load() or rebuild().
     *  @param src    Path where the package directory is stored.
     *  @param parent Parent package of this package
     */
    public PackageRepository(File src, PackageDecl parent) {
        super(src);
        _parent = parent;
    }

    /** Create a new repository for a package in a given directory. The
     *  contents of the repository will have to be loaded using the
     *  method load() or rebuild().
     *  @param src Path where the package directory is stored.
     *  @param parent Parent package of this package
     */
    public PackageRepository(String src, PackageDecl parent) {
        this(new File(src),parent);
    }


    ///////////////////////////////////////////////////////////////////
    ////                          public methods                   ////

    /** Load the repository from disk. Load the contents of this repository
     *  (scope, set of member files, and mappings between declarations
     *  and source files) in sequential order. */
    public void load() {
        super.load();
        System.out.println("Calling load (" + _src + ")");
        try {
            FileInputStream in = new FileInputStream(_rep);
            ObjectInputStream s = new ObjectInputStream(in);
            _pkgDecl        = (PackageDecl) s.readObject();
            _allMemberFiles = (Set)         s.readObject();
            _filesByDecl    = (Hashtable)   s.readObject();
            _declsByFile    = (Hashtable)   s.readObject();
        } catch (Exception e) {
            throw new RuntimeException("Failed to load package repository"
                    + " of directory " + _src + ": " + e.toString());
        }
        _pkgDecl.setContainer(_parent);
        Scope scope  = _pkgDecl.initGetScope();
        Scope parent = (_parent == null ? null : _parent.initGetScope());
        scope.setParent(parent);
        System.out.println("Finished loading");
    }


    /** Store this repository to disk. Store the contents of this
     *  repository (scope, set of member files, and mappings between
     *  declarations and source files).
     *  @exception RuntimeException if the repository has not been loaded.
     */
    public void store() {
        super.store();
        // Avoid crossed references among packages
        Scope scope = _pkgDecl.initGetScope();

        scope.setParent(null);
        _pkgDecl.setContainer(null);
        try {
            FileOutputStream out = new FileOutputStream(_rep);
            ObjectOutputStream s = new ObjectOutputStream(out);
            s.writeObject(_pkgDecl);
            s.writeObject(_allMemberFiles);
            s.writeObject(_filesByDecl);
            s.writeObject(_declsByFile);
            s.flush();
        } catch (IOException e) {
            throw new RuntimeException("I/O error when storing file "
                    + _rep);
        } catch(Exception e) {
            throw new RuntimeException("Failed to store package repository"
                    + " of directory " + _src);
        }
        scope.setParent(_parent.initGetScope());
        _pkgDecl.setContainer(_parent);
    }

    /** Modify this loaded repository to bring it up-to-date. Re-use
     *  as much information as possible from previous compilations.
     *  Use SyntaxRepository to update the information about
     *  compile units.
     *  @exception RuntimeException if the repository has not been loaded.
     */
    public void update() {
        super.update();

        // Update deleted files
        Iterator iter = _allMemberFiles.iterator();
        while (iter.hasNext()) {
            String name = (String) iter.next();
            File file = new File(_src, name);
            if (!file.exists()) {
                // File has been deleted
                // Check if it was a subpackage or a compile unit
                if (_declsByFile.contains(name))
                    removeSubPackage(name);
                else
                    removeCompileUnit(name);
            }
        }

        // Update modified or added files
        File[] files = _src.listFiles();
        for (int i = 0; i < files.length; i++) {
            File file = files[i];
            if (file.lastModified() < _rep.lastModified()) continue;
            String name = file.getName();
            boolean isCompileUnit = _declsByFile.contains(name);
            if (_allMemberFiles.contains(name)) {
                // File has been modified
                if (isCompileUnit)
                    removeCompileUnit(name);
                else
                    removeSubPackage(name);
            }
            // File has been modified or added
            if (isCompileUnit)
                buildCompileUnit(file);
            else
                buildSubPackage(file);
        }
    }

    /** Rebuild the information in this repository completely from
     *  scratch. Parse all source files, and create a new package
     *  member for each subdirectory in this directory.
     */
    public void rebuild() {
        super.rebuild();


        // Initialize data structures
        Scope parentScope = (_parent == null ? null : _parent.initGetScope());
        Scope scope       = new Scope(parentScope);
        _pkgDecl          = new PackageDecl(_src.getName(),_parent);
        _pkgDecl.setScope(scope);
        _allMemberFiles   = new HashSet();
        _filesByDecl      = new Hashtable();
        _declsByFile      = new Hashtable();

        // FIXME: Comment for debug
        System.out.println("Call rebuild( "  + _pkgDecl.fullName() + ")");

        // Parse all compile units
        // Build a package declaration for all subdirectories
        File[] files = _src.listFiles();
        for (int i = 0; i < files.length; i++) {
            File file = files[i];
            if (!isSource(file) && !file.isDirectory()) continue;
            _allMemberFiles.add(file.getName());
            if (file.isFile()) buildCompileUnit(file);
            else               buildSubPackage(file);
        }
    }

    /** Get the declaration of the package stored in this repository.
     *  @param parentPkg Parent package of the package contained in
     *  this repository.
     *  @return The package declaration contained in this repository.
     */
    public PackageDecl getPackageDecl(PackageDecl parentPkg ) {
        // If we have created a dummy declaration in the top level
        // package, we have to store all the information of the
        // package in the dummy declaration. Otherwise, we can use
        // _pkgDecl as the declaration.

        // FIXME: Comment added for debbugging purposes
        System.out.println("Package " + parentPkg.fullName() + "."
                + _pkgDecl.getName() + " is ready!");

        String name = _pkgDecl.getName();
        PackageDecl pkg = null;
        if (parentPkg._membersLoaded)
            pkg = parentPkg.getSubPackage(name);
        if (pkg == null) {
            pkg = _pkgDecl;
            Scope scope = parentPkg.initGetScope();
            scope.add(pkg);
        } else {
            Scope parentScope = parentPkg.initGetScope();
            Scope childScope  = _pkgDecl.initGetScope();
            // Set the scope of the dummy declaration
            pkg.getScope().copyDeclList(childScope);
            pkg.getScope().setParent(parentScope);
            // Make all class declarations and subpackages
            // refer to the dummy package declaration in
            // its 'container' field.
            Iterator iter = pkg.getScope().allLocalDecls();
            while (iter.hasNext()) {
                MetaModelDecl decl = (MetaModelDecl) iter.next();
                if (decl.hasContainer() &&
                        (decl.getContainer() == _pkgDecl))
                    decl.setContainer(pkg);
            }
        }
        pkg._membersLoaded = true;
        return pkg;
    }

    /** Get the name of file where an object declaration is defined.
     *  @param object Declaration of the object that we are trying
     *  to locate.
     *  @return The name of the file where an object declaration is
     *  defined, or null if it is not defined in this package.
     */
    public String sourceOfDecl(ObjectDecl object) {
        return (String) _filesByDecl.get(object);
    }

    /** Get the array of object declarations that are defined inside a
     *  file of this package.
     *  @param filename  The name of the file.
     *  @return The array of declarations of objects of the file, or
     *  null if the file does not belong to this package.
     */
    public ObjectDecl[] declsOfSource(String filename) {
        return (ObjectDecl[]) _declsByFile.get(filename);
    }


    ///////////////////////////////////////////////////////////////////
    ////                        protected variables                ////

    /** Package declaration representing the current package. When it is
     *  stored to a file, the following limitations apply: the pointer
     *  to the container is set to null, and the pointer to the parent
     *  scope in the scope of the package is also set to null. This
     *  package declaration is a dummy declaration, and it is not the
     *  declaration that will be used. That one will one the singleton
     *  declaration for this package.
     */
    protected PackageDecl _pkgDecl = null;

    /** Set of source file names and subdirectory names of this package.
     *  This is used to test if any file in the package has been modified,
     *  added or deleted.
     */
    protected Set _allMemberFiles = null;

    /** Mapping from ObjectDecls to the filenames where they are declared.
     *  Used to get the file where an ObjectDecl is declared.
     */
    protected Hashtable _filesByDecl = null;

    /** Mapping from a file name to the array of ObjectDecls declared in
     *  that file. This file is the inverse mapping of _filesByDecl.
     */
    protected Hashtable _declsByFile = null;

    /** Parent package of this package.
     */
    protected PackageDecl _parent = null;


    ///////////////////////////////////////////////////////////////////
    ////                         protected methods                 ////

    /** Check if this repository exists and is completely up-to-date.
     *  @return true iff the repository exists and the sources have
     *  not changed since the repository was created.
     */
    protected boolean isValid() {
        // If the repository is not loaded, we cannot know if a file
        // has been deleted. Therefore, we assume that it is not
        // valid and it has to be loaded.
        System.out.println("Checking if repository is valid");
        if (!_loaded) load();

        System.out.println("Package is loaded: check newer files");

        if (!super.isValid()) return false;

        System.out.println("No new/modified files: check deleted files");
        // Check that files have not been deleted.
        Iterator iter = _allMemberFiles.iterator();
        while (iter.hasNext()) {
            String name = (String) iter.next();
            File file = new File (_src, name);
            if (!file.exists()) return false;
        }
        return true;
    }

    /** Build all information relative to a compile unit. Use the
     *  SyntaxRepository of the compile unit.
     *  @param file Compile unit.
     */
    protected void buildCompileUnit(File file) {
        // Load this file and add it to the cache
        SyntaxRepository syntax = new SyntaxRepository(file);
        syntax.loadContent();
        ObjectDecl[] objects = syntax.getObjectDecls();
        // Update mappings for this file
        String name = file.getName();
        _declsByFile.put(name, objects);
        for (int i = 0; i < objects.length; i++) {
            _filesByDecl.put(objects[i], name);
        }
    }

    /** Build a subpackage of this package.
     *  @param dir Directory where the subpackage is stored.
     */
    protected void buildSubPackage(File dir) {
        PackageDecl subpkg = new PackageDecl(dir.getName(), _pkgDecl);
        //FIXME: Comment added for debugging
        System.out.println("Parent package is "+ _pkgDecl.fullName());
        System.out.println("Package " + subpkg.fullName() + " created");

        Scope parent = _pkgDecl.initGetScope();
        parent.add(subpkg);
        System.out.println("Adding package"+ subpkg.fullName() + "to parent");

        // FIXME: subpkg.setScope(null) ?
        subpkg.setScope(new Scope(parent));
        //System.out.println("Package " + subpkg.getScope());

        //System.out.println("Scope of package " + subpkg.fullName() + "is" +
        //                   subpkg.getScope());
    }

    /** Remove from this repository information about a file that
     *  has been deleted or modified. The set of file members in
     *  the directory does not have to be updated.
     *  @param name Name of the file that has to be removed.
     */
    protected void removeCompileUnit(String name) {
        ObjectDecl[] objects = (ObjectDecl[]) _declsByFile.get(name);
        _declsByFile.remove(name);
        for (int i = 0; i < objects.length; i++) {
            ObjectDecl object = objects[i];
            _filesByDecl.remove(object);
            _pkgDecl.initGetScope().remove(object);
        }
    }

    /** Remove from this repository information about a subpackage
     *  that has been deleted or modified. The set of file members in
     *  the directory does not have to be updated.
     *  @param name Name of the subpackage that has to be removed.
     */
    protected void removeSubPackage(String name) {
        Scope scope        = _pkgDecl.initGetScope();
        PackageDecl subpkg = (PackageDecl)
            scope.lookupLocal(name, CG_PACKAGE);
        _pkgDecl.initGetScope().remove(subpkg);
    }


    /** Get the name of the repository file.
     *  @return The name of the repository file.
     */
    protected String _getRepositoryName() { return ".package"; }

}



